%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% ======================== Bayesian Inversion ========================== %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%                                                                        %
%  04.03.2022                                                            %
%                                                                        %
%  Paul Scherrer Institut                                                %
%  David Breitenmoser                                                    %
%  OFLD/U06                                                              %
%  Forschungsstrasse 111                                                 %
%  5232 Villigen PSI                                                     %
%  Switzerland                                                           %
%                                                                        %
%  Phone: +41 56 310 44 61                                               %
%  E-Mail: david.breitenmoser@psi.ch                                     %
%                                                                        %
%                                                                        %
%                                                                        %
%  Description:                                                          %
%    Perform Bayesian Inversion                                          %
%                                                                        %
%                                                                        %
% Copyright (c) 2022, David Breitenmoser                                 %
% All rights reserved.                                                   %
%                                                                        %
% Redistribution and use in source and binary forms, with or without     %
% modification, are permitted provided that the following conditions are %
% met:                                                                   %
%                                                                        %
%     * Redistributions of source code must retain the above copyright   %
%       notice, this list of conditions and the following disclaimer.    %
%     * Redistributions in binary form must reproduce the above copyright%
%       notice, this list of conditions and the following disclaimer in  %
%       the documentation and/or other materials provided with the       %
%       distribution                                                     %
%                                                                        %
% THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS%
% IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED  %
% TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A        %
% PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT     %
% OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,  %
% SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT       %
% LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,  %
% DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY  %
% THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT    %
% (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE  %
% OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.   %
%                                                                        %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%----------------------------- Preparation ----------------------------%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%% --- tabula rasa ---
close all
clear variables
fclose('all');
clc

%% --- Change Default Values ---
set(groot,'defaulttextinterpreter','latex')
set(groot, 'defaultAxesTickLabelInterpreter','latex');
set(groot, 'defaultLegendInterpreter','latex');

set(0,'DefaultTextFontname', 'Latin Modern Roman')
set(0,'DefaultAxesFontName', 'Latin Modern Roman')
set(0,'DefaultLegendFontName','Latin Modern Roman');

set(0, 'DefaultLineLineWidth', 1);

%% --- Set random number generator for reproducable results ---
rng(100,'twister')

%% --- Initilaize UQlab ---
uqlab

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%---------------------------- Input Deck ------------------------------%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


%% 1) Input Folder
main_path=['J:\Datenanalyse\Intrinsic Resolution\Det4_October_2021\Bayesian_Approach'];
ProbMdl_path=[main_path '\ProbModel\ProbModel_v3.mat'];
ExpData_path=[main_path '\FLUKA\FLUKA_ExpDesign_v3_v1\Postprocessing\exp\v2\SPEC_response.mat'];
MetaModel_path=[main_path '\Emulator\FLUKA_ExpDesign_v3_v1\GPR\v2\GPR_model.mat'];

%% 2) Output Folder
output_path=[main_path '\Bayesian Inversion\FLUKA_ExpDesign_v3_v1\test_v2'];
CHECKfolder_v1(output_path)

%% 3) Output Normalization
norm=10^-6;

%% 4) Solver Options
N_chains=0.3*10^2;                                                           % [-] Number of chains
N_steps=1*10^3;                                                             % [-] Number of steps

%% 5) Postprocessing Options
N_prior=10^3;                                                               % [-] Number of prior samples
N_priorp=10^3;                                                              % [-] Number of prior predictive samples
N_postp=10^3;                                                               % [-] Number of posterior predictive samples


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%-------------------------------- Load Data ---------------------------%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% %% 1) Load Probabilistic Input Model
% disp(['Load Probabilistic Input Model , ' char(datetime)]);
% load([ProbMdl_path])
% N_var=length(ProbModel.nonConst);

%% 2) Load Experimental Data
disp(['Load Experimental Data , ' char(datetime)]);
load([ExpData_path],'Y')
N_det=size(Y.mean_exp.value,1);

%% 3) Load Metamodel
disp(['Load Metamodel , ' char(datetime)]);
load([MetaModel_path],'myGPR')

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%-------------------- Define Probabilistic Input Model ----------------%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%% 1) Command Line Output
disp(['Define Probabilistic Input Model , ' char(datetime)]);

%% 2) Define Probabilistic Model

% i) Birks coefficient
Input.Marginals(1).Type = 'Uniform';
Input.Marginals(1).Parameters = [100 500];                                  % [MeV/cm]
Input.Marginals(1).Name='co_brk';

% ii) Onsager coefficient
Input.Marginals(2).Type = 'Constant';
Input.Marginals(2).Parameters = 36.4;                                       % [MeV/cm]
Input.Marginals(2).Name='co_ons';

% iii) Free electron/hole formation efficiency coefficient
Input.Marginals(3).Type = 'Uniform';
Input.Marginals(3).Parameters = [0.4 0.58];                                 % [-]
Input.Marginals(3).Name='co_eff';

% iv) Debye-related trapping coefficient
Input.Marginals(4).Type = 'Uniform';
Input.Marginals(4).Parameters = [1 20];                                    % [MeV/cm]
Input.Marginals(4).Name='co_tr';
Input.Name='ProbModel_v1';
Input.Copula.Type = 'Independent';
% Input.Marginals(4).Type = 'constant';
% Input.Marginals(4).Parameters = [12];                                    % [MeV/cm]
% Input.Marginals(4).Name='co_tr';
% Input.Name='ProbModel_v1';
% Input.Copula.Type = 'Independent';

%% 3) Create Probabilistic Model
ProbModel = uq_createInput(Input);
N_var=length(ProbModel.nonConst);


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%---------------------- Perform Bayesian Inversion --------------------%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%% 1) Setup Bayesian Inversion Module
BayesOpts.Type = 'Inversion';
BayesOpts.Name=['BayesianInversion_det' num2str(round(1-1))];

%% 2) Experimental Data
BayesOpts.Data.y = Y.mean_exp.value{1,1}'./norm;
myData=BayesOpts.Data.y./norm;
N_y=length(Y.mean_exp.value{1,1});

%% 3) Forward Model
modelopts.mFile='uq_PCA_norm_evalModel';
modelopts.Parameters=myGPR{1,1};
myMeta=myGPR{1,1};
uq_myMdl=uq_createModel(modelopts);
BayesOpts.ForwardModel.Model = uq_myMdl;

%% 4) Prior Model
BayesOpts.Prior = ProbModel;

%% 5) Discrepancy Model

% i) Discrepancy model type
BayesOpts.Discrepancy.Type = 'Gaussian';

% ii) Discrepancy prior model
DiscrepancyPriorOpts.Name=['Discrepancy_Prior'];
DiscrepancyPriorOpts.Marginals.Type='Exponential';
DiscrepancyPriorOpts.Marginals.Moments=median((10^0.*Y.mean_exp.value{1,1}./norm).^2);
%     DiscrepancyPriorOpts.Name='Discrepancy_Prior';
%     DiscrepancyPriorOpts.Marginals.Type='Uniform';
%     DiscrepancyPriorOpts.Marginals.Parameters=[0 median((10^-3.*Y.mean_exp.value{1,1}./norm).^2)];
myDiscrepancyPrior= uq_createInput(DiscrepancyPriorOpts);
BayesOpts.Discrepancy.Prior = myDiscrepancyPrior;

%% 6) Solver
BayesOpts.Solver.Type = 'MCMC';
BayesOpts.Solver.MCMC.Sampler = 'AIES';
BayesOpts.Solver.MCMC.NChains = N_chains;
BayesOpts.Solver.MCMC.Steps = N_steps;
BayesOpts.Solver.MCMC.a=2;

BayesOpts.Solver.MCMC.Visualize.Parameters = [1:N_var+1];
BayesOpts.Solver.MCMC.Visualize.Interval = round(0.05*BayesOpts.Solver.MCMC.Steps);

%% 7) Execute Inversion
myBayesianAnalysis = uq_createAnalysis(BayesOpts);

%% 8) Postprocessing

% i) Display main results
uq_print(myBayesianAnalysis)

uq_display(myBayesianAnalysis,'trace',[1:N_var])
uq_display(myBayesianAnalysis,'scatterplot',[1:N_var])
uq_display(myBayesianAnalysis,'meanConvergence',[1:N_var])
uq_display(myBayesianAnalysis,'acceptance',true)

% ii) Define bad chains
badChainsIndex1 = squeeze(myBayesianAnalysis.Results.Sample(end,1,:) < 300)';
badChainsIndex2 = squeeze(myBayesianAnalysis.Results.Sample(end,2,:) > 0.56)';
badChainsIndex3 = squeeze(myBayesianAnalysis.Results.Acceptance < 0.25);
badChainsIndex=any([badChainsIndex1 ; badChainsIndex2 ; badChainsIndex3]);

% iii) Postprocessing Script
uq_postProcessInversionMCMC(myBayesianAnalysis,...
    'burnIn', 0.5,...
    'percentiles',[0.05,0.95],...
    'badChains',badChainsIndex,...
    'pointestimate',{'MAP'},...
    'gelmanRubin',true,...                                              
    'prior',N_prior,...                                                     % Number of prior (parameter value) samples | N_prior x M
    'priorPredictive',N_priorp,...                                          % Number of prior predictive (output value) samples | N_priorp x N_out
    'posteriorPredictive',N_postp);                                         % Number of posterior predictive (output value) samples | N_postp x N_out

% iv) Evaluate prior, likelihood and posterior PDF
N_stepp=size(myBayesianAnalysis.Results.PostProc.PostSample,1);
N_varp=size(myBayesianAnalysis.Results.PostProc.PostSample,2);
N_chainp=size(myBayesianAnalysis.Results.PostProc.PostSample,3);
% PostSample_reshaped=squeeze(reshape(myBayesianAnalysis.Results.PostProc.PostSample,[N_chainp*N_stepp,N_varp]));
PostSample_reshaped=squeeze(myBayesianAnalysis.Results.PostProc.PostSample(:,:,1));

PDF_eval.Prior_eval=uq_evalPDF(PostSample_reshaped,myBayesianAnalysis.Internal.FullPrior);
PDF_eval.LogPrior_eval=uq_evalLogPDF(PostSample_reshaped,myBayesianAnalysis.Internal.FullPrior);

PDF_eval.Likeli_eval=uq_inversion_likelihood(PostSample_reshaped,myBayesianAnalysis.Internal,'Likelihood');
PDF_eval.LogLikeli_eval=uq_inversion_likelihood(PostSample_reshaped,myBayesianAnalysis.Internal,'LogLikelihood');

PDF_eval.Posterior_eval_unnom=PDF_eval.Likeli_eval.*PDF_eval.Prior_eval;
PDF_eval.LogPosterior_eval_unnom=PDF_eval.LogPrior_eval+PDF_eval.LogLikeli_eval;

PDF_eval.Posterior_eval_nom=PDF_eval.Posterior_eval_unnom./sum(PDF_eval.Posterior_eval_unnom);
PDF_eval.LogPosterior_eval_nom=PDF_eval.LogPosterior_eval_unnom./sum(PDF_eval.LogPosterior_eval_unnom);

%% 9) Response Variable Plot
figure
plot(myBayesianAnalysis.Data.y,'k-')
hold on
plot(myBayesianAnalysis.Results.PostProc.PointEstimate.ForwardRun{1}.Out,'r-')
hold on
post_quant_sample=quantile(myBayesianAnalysis.Results.PostProc.PostPredSample.Sample,[0.05 0.95],1);
post_quant_model=quantile(myBayesianAnalysis.Results.PostProc.PostPredSample.ModelEvaluations,[0.05 0.95],1);
plot(post_quant_sample(1,:),'--r')
plot(post_quant_sample(2,:),'--r')
plot(post_quant_model(1,:),':r')
plot(post_quant_model(2,:),':r')


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%----------------------------- Save Data  -----------------------------%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

disp(['Save Results , ' char(datetime)]);

save([output_path '\ProbModel.mat'],...
    'ProbModel',...
    '-v7.3')

save([output_path '\myMdl.mat'],...
    'uq_myMdl',...
    '-v7.3')

save([output_path '\myData.mat'],...
    'myData',...
    '-v7.3')

save([output_path '\BayesOpts.mat'],...
    'BayesOpts',...
    '-v7.3')

save([output_path '\myMeta.mat'],...
    'myMeta',...
    '-v7.3')

save([output_path '\PDF_eval.mat'],...
    'PDF_eval',...
    '-v7.3')

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%------------------------ Metamodel Plots  ----------------------------%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


%% 1) X1 parameter
figure
plot(Y.mean_exp.value{1,1}'./norm,'k')
hold on
par={[100 36.4 0.6 12],[200 36.4 0.6 12],[300 36.4 0.6 12],[400 36.4 0.6 12]};
N_sim=length(par);
legend_lbl=cell(N_sim+1,1);
legend_lbl{1}='Experimental data';
for k=1:N_sim
    plot(uq_evalModel(par{k}))
    legend_lbl{k+1}=['Model' num2str(k)];
end
legend(legend_lbl)
xlabel('Output index')
ylabel('Y [CPS $\mathrm{Bq^{-1}}$]')

%% 2) X2 parameter
figure
plot(Y.mean_exp.value{1,1}'./norm,'k')
hold on
par={[200 36.4 0.4 12],[200 36.4 0.5 12],[200 36.4 0.6 12],[200 36.4 0.65 12]};
N_sim=length(par);
legend_lbl=cell(N_sim+1,1);
legend_lbl{1}='Experimental data';
for k=1:N_sim
    plot(uq_evalModel(par{k}))
    legend_lbl{k+1}=['Model' num2str(k)];
end
legend(legend_lbl)
xlabel('Output index')
ylabel('Y [CPS $\mathrm{Bq^{-1}}$]')

%% 3) X3 parameter
figure
plot(Y.mean_exp.value{1,1}'./norm,'k')
hold on
par={[200 36.4 0.6 10],[200 36.4 0.6 12],[200 36.4 0.6 14],[200 36.4 0.6 16]};
N_sim=length(par);
legend_lbl=cell(N_sim+1,1);
legend_lbl{1}='Experimental data';
for k=1:N_sim
    plot(uq_evalModel(par{k}))
    legend_lbl{k+1}=['Model' num2str(k)];
end
legend(legend_lbl)
xlabel('Output index')
ylabel('Y [CPS $\mathrm{Bq^{-1}}$]')

