function [YPC,YPC_var,YPC_replications]=uq_PCA_norm_evalModel(X,myMdl)
%uq_PCA_evalModel Evaluates UQLab model for PCA transformed inputs       %
%                                                                        %
%                                                                        %
% Input                                                                  %
%   X:                     PCA transformated X" matrix | N" x K'         %
%   myMdl:                 Struct containing custom PCA and UQLab model  %
%    myMdl.model:          UQLab model                                   %
%    myMdl.PCA:            PCA model                                     %
%     myMdl.PCA.T:         PCA transformation matrix | K x K'            %
%     myMdl.PCA.Z:         PCA transformated X matrix | N x K'           %
%     myMdl.PCA.mean:      mean X vector | 1 x K                         %
%     myMdl.PCA.std:       sample standard deviation X vector | 1 x K    %
%     myMdl.PCA.threshold: PCA threshold | 1 x 1                         %
%     myMdl.PCA.X:         data matrix to create PCA model | N x K       %
%    myMdl.rMSE:           relative mean squared error (validation)      %
%                          | 1 x 1                                       %
%                                                                        %
% Output:                                                                %
%   YPC:                   predicted response in original data space     %
%                          | N" x K                                      %
%   YPC_var:               sample variance of the predictions at each    % 
%                          data point and for each  dimension| N" x K    %
%   YPC_replications:      bPCE predictions for each of the B bootstrap  %
%                          replications | N" x B x K                     %
%                          (only for PCE models)                         %
%                                                                        %
%                                                                        %
% Notes:                                                                 %
%   - N:  # data points in input data matrix X used to create the        %
%           PCA model [1,inf)                                            %
%     N": # data points in input data matrix X" [1,inf)                  %
%     K:  # variables in input data matrix X [1,inf)                     %
%     K': # variables in PCA reduced data space ( K' <= K )              %
%                                                                        %
%                                                                        %
% Written by:                                                            %
%  Paul Scherrer Institut                                                %
%  David Breitenmoser                                                    %
%  OFLD/U06                                                              %
%  Forschungsstrasse 111                                                 %
%  5232 Villigen PSI                                                     %
%  Switzerland                                                           %
%                                                                        %
%  Phone: +41 56 310 44 61                                               %
%  E-Mail: david.breitenmoser@psi.ch                                     %
%                                                                        %                                                                        %
%                                                                        %
% Copyright (c) 2022, David Breitenmoser                                 %
% All rights reserved.                                                   %
%                                                                        %
% Redistribution and use in source and binary forms, with or without     %
% modification, are permitted provided that the following conditions are %
% met:                                                                   %
%                                                                        %
%     * Redistributions of source code must retain the above copyright   %
%       notice, this list of conditions and the following disclaimer.    %
%     * Redistributions in binary form must reproduce the above copyright%
%       notice, this list of conditions and the following disclaimer in  %
%       the documentation and/or other materials provided with the       %
%       distribution                                                     %
%                                                                        %
% THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS%
% IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED  %
% TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A        %
% PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT     %
% OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,  %
% SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT       %
% LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,  %
% DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY  %
% THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT    %
% (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE  %
% OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.   %
%                                                                        %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


%% 1) Process input parameters

if nargin > 2
    disp('ERROR: Too many inputs')
    return    
elseif nargin < 2
    disp('ERROR: Too few inputs')
    return
end

norm=10^-6;

if strcmp(myMdl.Type,'SSE') || strcmp(myMdl.Type,'SVR')
    
    %% 3) Perform UQ Model Evaluation in PCA reduced data space
    [ZPC]=uq_evalModel(myMdl.model,X);
    
    %% 4) Transform back to original data space
    [YPC]=PCA_Z2X(ZPC,myMdl.PCA)./norm;

% elseif strcmp(myMdl.Type,'PCK') 
%     
%     %% 3) Perform UQ Model Evaluation in PCA reduced data space
%     [ZPC,ZPC_var]=uq_evalModel(myMdl.model,X);
%     
%     %% 4) Transform back to original data space
%     [YPC]=PCA_Z2X(ZPC,myMdl.PCA);
%     [YPC_var]=(PCA_Z2X(sqrt(ZPC_var)+ZPC,myMdl.PCA)-YPC).^2;
    
else
    
    %% 3) Perform UQ Model Evaluation in PCA reduced data space
    [ZPC,ZPC_var,ZPC_replications]=uq_evalModel(myMdl.model,X);
    
    %% 4) Transform back to original data space
    [YPC]=PCA_Z2X(ZPC,myMdl.PCA)./norm;
    [YPC_var]=(PCA_Z2X(sqrt(ZPC_var)+ZPC,myMdl.PCA)-YPC).^2./norm.^2;
    [YPC_replications]=PCA_Z2X(ZPC_replications,myMdl.PCA)./norm;
    
    
end

